import glob, os
import pandas as pd
import numpy as np
from sklearn.metrics import roc_auc_score

def read_csv_(path, filename):

    file = glob.glob(os.path.join(path, filename))
    short_file = []
    long_file = []
    
    for file_ele in file:
        if len(file_ele) > 43:
            long_file.append(file_ele)
        else:
            short_file.append(file_ele)
    
    short_file = sorted(short_file, key=lambda file_ele: file_ele[-5])    
    long_file = sorted(long_file, key=lambda file_ele: file_ele[-5])     
    file =  short_file + long_file
    print(file)
        
    dl = []
    for f in file:
        data = pd.read_csv(f, header=None)
        dl.append(data)
    df = pd.concat(dl)   
    label_list = df.iloc[:, 0].tolist()  
    return label_list

path = r'./LOO-result'
filename_pre = "predict_label_first*.csv"
filename_true = "true_label_first*.csv"

filename_pre_second = "predict_label_second*.csv"
filename_true_second = "true_label_second*.csv"

Auc_score_first = "AUC_first_score*.csv"
Auc_score_second = "AUC_second_score*.csv"

Auc_true_first = "AUC_first_true*.csv"
Auc_true_second = "AUC_second_true*.csv"


predict_list = read_csv_(path, filename_pre)
true_list = read_csv_(path, filename_true)

predict_list_second = read_csv_(path, filename_pre_second)
true_list_second = read_csv_(path, filename_true_second)

Auc_score_first = read_csv_(path, Auc_score_first)
Auc_score_second = read_csv_(path, Auc_score_second)

Auc_true_first = read_csv_(path, Auc_true_first)
Auc_true_second = read_csv_(path, Auc_true_second)


# Acc calculated by the first layer model
initial_positive = 0  
predict_right = 0  
predict_wrong = 0  

initial_negtive = 0  
predict_negtive = 0 
predict_right_negtive = 0  

for i in range(len(predict_list)):
    predict_tmp = predict_list[i] 
    initial_tmp = true_list[i]  

    if int(initial_tmp) == 1:
        initial_positive += 1 
    if int(predict_tmp) == 1 and int(initial_tmp) == 1:
        predict_right += 1  
    if int(predict_tmp) == 0 and int(initial_tmp) == 1:
        predict_wrong += 1  

    if int(initial_tmp) == 0:
        initial_negtive += 1  
    if int(predict_tmp) == 1 and int(initial_tmp) == 0:
        predict_negtive += 1 
    if int(predict_tmp) == 0 and int(initial_tmp) == 0:
        predict_right_negtive += 1  

print('Test ACC:', (predict_right + predict_right_negtive) / (
            predict_right + predict_negtive + predict_right_negtive + predict_wrong))

# Acc calculated by the second layer model
initial_positive = 0  
predict_right = 0  
predict_wrong = 0  

initial_negtive = 0  
predict_negtive = 0  
predict_right_negtive = 0 

for i in range(len(predict_list_second)):
    predict_tmp = predict_list_second[i] 
    initial_tmp = true_list_second[i]  

    if int(initial_tmp) == 1:
        initial_positive += 1  
    if int(predict_tmp) == 1 and int(initial_tmp) == 1:
        predict_right += 1  
    if int(predict_tmp) == 0 and int(initial_tmp) == 1:
        predict_wrong += 1  

    if int(initial_tmp) == 0:
        initial_negtive += 1  
    if int(predict_tmp) == 1 and int(initial_tmp) == 0:
        predict_negtive += 1  
    if int(predict_tmp) == 0 and int(initial_tmp) == 0:
        predict_right_negtive += 1  

print('second Test ACC:', (predict_right + predict_right_negtive) / (
            predict_right + predict_negtive + predict_right_negtive + predict_wrong))

# Sn calculated by the first layer model
initial_positive = 0  
predict_right = 0 
predict_wrong = 0
for i in range(len(predict_list)):
    predict_tmp = predict_list[i]  
    initial_tmp = true_list[i]  

    if int(initial_tmp) == 1:
        initial_positive += 1  
    if int(predict_tmp) == 1 and int(initial_tmp) == 1:
        predict_right += 1 
    if int(predict_tmp) == 0 and int(initial_tmp) == 1:
        predict_wrong += 1 

print('Test Sn:', predict_right / (predict_right + predict_wrong))

# Sn calculated by the second layer model
initial_positive = 0  
predict_right = 0  
predict_wrong = 0
for i in range(len(predict_list_second)):
    predict_tmp = predict_list_second[i] 
    initial_tmp = true_list_second[i]  

    if int(initial_tmp) == 1:
        initial_positive += 1  
    if int(predict_tmp) == 1 and int(initial_tmp) == 1:
        predict_right += 1  
    if int(predict_tmp) == 0 and int(initial_tmp) == 1:
        predict_wrong += 1  

print('second Test Sn:', predict_right / (predict_right + predict_wrong))

# Sp calculated by the first layer model
initial_negtive = 0  
predict_negtive = 0  
predict_right_negtive = 0  

for i in range(len(predict_list)):
    predict_tmp = predict_list[i]  
    initial_tmp = true_list[i] 

    if int(initial_tmp) == 0:
        initial_negtive += 1  
    if int(predict_tmp) == 1 and int(initial_tmp) == 0:
        predict_negtive += 1  
    if int(predict_tmp) == 0 and int(initial_tmp) == 0:
        predict_right_negtive += 1 

print('Test Sp:', predict_right_negtive / (predict_right_negtive + predict_negtive))

# Sp calculated by the second layer model
initial_negtive = 0  
predict_negtive = 0  
predict_right_negtive = 0  

for i in range(len(predict_list_second)):
    predict_tmp = predict_list_second[i] 
    initial_tmp = true_list_second[i]  

    if int(initial_tmp) == 0:
        initial_negtive += 1  
    if int(predict_tmp) == 1 and int(initial_tmp) == 0:
        predict_negtive += 1  
    if int(predict_tmp) == 0 and int(initial_tmp) == 0:
        predict_right_negtive += 1  

print('second Test Sp:', predict_right_negtive / (predict_right_negtive + predict_negtive))

# Mcc calculated by the first layer model
initial_positive = 0 
predict_right = 0  
predict_wrong = 0   

initial_negtive = 0 
predict_negtive = 0  
predict_right_negtive = 0   

for i in range(len(predict_list)):
  predict_tmp = predict_list[i]  
  initial_tmp = true_list[i]  

  if int(initial_tmp) == 1:
    initial_positive += 1    
  if int(predict_tmp) == 1 and int(initial_tmp) == 1:
    predict_right += 1    
  if int(predict_tmp) == 0 and int(initial_tmp) == 1:
    predict_wrong += 1  

  if int(initial_tmp) == 0:
    initial_negtive += 1    
  if int(predict_tmp) == 1 and int(initial_tmp) == 0:
    predict_negtive += 1    
  if int(predict_tmp) == 0 and int(initial_tmp) == 0:
    predict_right_negtive += 1    

fenzi = (predict_right*predict_right_negtive) - (predict_negtive*predict_wrong)
fenmu = ((predict_right+predict_wrong)*(predict_right+predict_wrong)*(predict_right_negtive+predict_negtive)*(predict_right_negtive+predict_wrong))**0.5
print('Test MCC:', fenzi/fenmu)

# Mcc calculated by the second layer model
initial_positive = 0 
predict_right = 0  
predict_wrong = 0   

initial_negtive = 0 
predict_negtive = 0  
predict_right_negtive = 0   

for i in range(len(predict_list_second)):
  predict_tmp = predict_list_second[i]  
  initial_tmp = true_list_second[i]  

  if int(initial_tmp) == 1:
    initial_positive += 1    
  if int(predict_tmp) == 1 and int(initial_tmp) == 1:
    predict_right += 1   
  if int(predict_tmp) == 0 and int(initial_tmp) == 1:
    predict_wrong += 1   

  if int(initial_tmp) == 0:
    initial_negtive += 1  
  if int(predict_tmp) == 1 and int(initial_tmp) == 0:
    predict_negtive += 1   
  if int(predict_tmp) == 0 and int(initial_tmp) == 0:
    predict_right_negtive += 1   

fenzi = (predict_right*predict_right_negtive) - (predict_negtive*predict_wrong)
fenmu = ((predict_right+predict_wrong)*(predict_right+predict_wrong)*(predict_right_negtive+predict_negtive)*(predict_right_negtive+predict_wrong))**0.5
print('second Test MCC:', fenzi/fenmu)

# AUC calculated by the first layer model
Test_y_true = np.array(Auc_true_first)
Test_y_scores = np.array(Auc_score_first)
print("Test AUC is", roc_auc_score(Test_y_true, Test_y_scores))

# AUC calculated by the second layer model
Test_y_true = np.array(Auc_true_second)
Test_y_scores = np.array(Auc_score_second)
print("second Test AUC is", roc_auc_score(Test_y_true, Test_y_scores))