from keras.layers import Multiply, BatchNormalization
from keras.layers.core import *
from keras.models import *
from keras.layers import Dense, Dropout, LSTM, Embedding, Bidirectional
from keras.preprocessing.sequence import pad_sequences
import os
import copy
import random
import pandas as pd
from keras.callbacks import ModelCheckpoint



# Hyperparameter settings
k_merList = [2]  # kmer
kkk = k_merList[0]
EPCOHS = 6 #  an arbitrary cutoff, generally defined as "one pass over the entire dataset", used to separate training into distinct phases, which is useful for logging and periodic evaluation.
BATCH_SIZE = 64 # a set of N samples. The samples in a batch are processed` independently,
                   #in parallel. If training, a batch results in only one update to the model.
maxlen = 718
RNN_HIDDEN_DIM = 64    # Number of neural units in the lstm layer
DROPOUT_RATIO_1 = 0.5 # proportion of neurones not used for training    
DROPOUT_RATIO_2 = 0.5 # proportion of neurones not used for training    
xSquare_selection = 45 # Filter out the number of SeqPose Feature

checkpoint_dir = './checkpoints'  # Store model files
if not os.path.exists(checkpoint_dir):   # if the path does not exist, generate it
    os.mkdir(checkpoint_dir)

checkpoint_dir = './checkpoints/final_model'  # Store model files
if not os.path.exists(checkpoint_dir):   # if the path does not exist, generate it
    os.mkdir(checkpoint_dir)

output_dir = './output'  # Store model files
if not os.path.exists(output_dir):  # if the path does not exist, generate it
    os.mkdir(output_dir)

useless_filename = './output/Useless_SeqPose_features_'+str(kkk)+'_mer_'+str(xSquare_selection)+'.csv'   # File storing sequence position
filepath = checkpoint_dir + "/first_layer_model_"+str(kkk)+"_mer_"+str(xSquare_selection)+".hdf5"  # First layer model
filepath_second = checkpoint_dir + "/second_layer_model_"+str(kkk)+"_mer_"+str(xSquare_selection)+".hdf5"  # Second layer model

# Read data
def read_Fasta():

    finalFasta_positive = {}  # Store positive samples
    finalFasta_positive_second = {}  # Under the premise of a positive sample, store the enhancer intensity
    fileList = ['./data/weak_enhancer.txt', './data/strong_enhancer.txt']  
    for m in fileList:
        fasta = {}  
        second_label = {}  
        if m == './data/strong_enhancer.txt':  
            with open(m) as file:
                sequence = ""
                for line in file:
                    if line.startswith(">"):
                        # Remove the \n and> in the description field line
                        name = line[1:].rstrip()  
                        fasta[name] = ''
                        continue
                    fasta[name] += line.rstrip()  
                    second_label[name + ' strong'] = 'strong'

                finalFasta_positive = dict(finalFasta_positive, **fasta)  # Combine all positive samples
                finalFasta_positive_second = dict(finalFasta_positive_second, **second_label) 
        else:
            with open(m) as file:  # Storage weak category label
                sequence = ""
                for line in file:
                    if line.startswith(">"):
                        name = line[1:].rstrip()
                        fasta[name] = ''
                        continue
                    # fasta[name] += line.rstrip().upper()          # 获得字典 （将所有的小写转为大写）
                    fasta[name] += line.rstrip()  
                    second_label[name + ' weak'] = 'weak'
                finalFasta_positive = dict(finalFasta_positive, **fasta)   # Combine all positive samples
                finalFasta_positive_second = dict(finalFasta_positive_second, **second_label)  

    finalFasta_negative = {}  # Store negative samples
    finalFasta_negative_second = {} 
    fasta = {} 
    second_label = {}  
    with open('./data/non-enhancer.txt') as file:
        sequence = ""
        for line in file:
            if line.startswith(">"):
                name = line[1:].rstrip()
                fasta[name] = ''
                continue
            fasta[name] += line.rstrip()  
            second_label[name + ' second label'] = 'non enhancer'
        finalFasta_negative = dict(finalFasta_negative, **fasta)  # Combine negative samples
        finalFasta_negative_second = dict(finalFasta_negative_second, **second_label)  

    AllPositive_items = list(finalFasta_positive.items())  # Get the key-value pair of the positive sample     
    AllNegative_items = list(finalFasta_negative.items())  # Get key-value pairs of negative samples

    AllStrength_items = list(finalFasta_positive_second.items())  

    sequenceList = []  

    for i in range(len(AllPositive_items)):
        tuple_Strength_key = AllStrength_items[i][0]
        tuple_Strength_val = AllStrength_items[i][1]

        tuple_sequence_key = AllPositive_items[i][0]
        tuple_sequence_val = AllPositive_items[i][1]
        if tuple_Strength_key[-2] == 'n':
            tmpDict = {'sequence': tuple_sequence_val, 'target': 1, 'strength': 1}  # Build the Strong Enhancer data set
            sequenceList.append(tmpDict)
        else:
            tmpDict = {'sequence': tuple_sequence_val, 'target': 1, 'strength': 0}  # Build Weak Enhancer data set
            sequenceList.append(tmpDict)
            
    for key, value in AllNegative_items:
        if value[-1] != 'r':

            tmpDict = {'sequence': value, 'target': 0,
                       'strength': 333}  # Build a non-enhancer dataset
            sequenceList.append(tmpDict)  
    return sequenceList

sequenceList = read_Fasta()
random.seed(75)
random.shuffle(sequenceList)  # shuffle

# Use kmer to segment raw data
def k_split(sequenceList, k_merList):

    def k_mer(sequence, k):
        
        # kmer algorithm
        splitSequence = []  # Used to store the segmented sequence
        for i in range(len(sequence) - k + 1):
            tmp = ''
            for j in range(i, i + k):
                tmp += sequence[j]

            splitSequence.append(tmp)
        return splitSequence

    finalSequenceList = []
    for k_merNum in k_merList:  
        sequenceTmpList = copy.deepcopy(sequenceList)  
        for i in range(len(sequenceList)):  
            sequenceTmpList[i]['sequence'] = k_mer(sequenceList[i]['sequence'],
                                                   k_merNum)  
            finalSequenceList.append(sequenceTmpList[i])

    # Take each word as a feature,
    avertFinal = finalSequenceList[:len(sequenceList)]
    for i in range(len(k_merList)):
        tmplist = finalSequenceList[len(sequenceList) * i:(i + 1) * len(sequenceList)]
        for dictindex in range(len(tmplist)):
            avertFinal[dictindex]['sequence'] = avertFinal[dictindex]['sequence'] + tmplist[dictindex]['sequence']

    finalSequenceList = avertFinal # finalSequenceList is a list, each element of which is a dictionary, and the sequence in finalSequenceList has been split by k-mer

    wordVectorSeqList = []  
    wordVectorTarList = []  
    wordVectorStrengthList = []  

    for allDictSeq in finalSequenceList:
        wordVectorSeqList.append(allDictSeq['sequence'])  
        wordVectorTarList.append(allDictSeq['target']) 
        wordVectorStrengthList.append(allDictSeq['strength'])

    return wordVectorSeqList, wordVectorTarList, wordVectorStrengthList

wordVectorSeqList, wordVectorTarList, wordVectorStrengthList = k_split(sequenceList, k_merList)

# generate SeqPoseDict
token_dict = {}
token_dict['UNK'] = 0
for pairs in wordVectorSeqList:
    for token in pairs:
        if token not in token_dict:
            token_dict[token] = len(token_dict)    # token_dict 是由词组成的字典，大致如下
SeqPoseDict_length = len(token_dict)

np.save('./output/SeqPoseDict.npy', token_dict)


# Serialization
def get_index(sentence):
    global token_dict   
    sequence = []
    for word in sentence:
        try:
            sequence.append(token_dict[word])
        except KeyError:
            sequence.append(0)
    return sequence

def feature_selection(X_data, filename, xSquare_selection):  # Feature selection

    sorted_index_df = pd.read_csv(filename, header=None)
    sorted_index_list = sorted_index_df.iloc[:, 0].tolist()  

    unuseful_column = []
    for i in range(xSquare_selection):
        unuseful_column.append(sorted_index_list[i])
    unuseful_column.sort(reverse=True)  

    for i in range(len(X_data)): 
        for column_ele in unuseful_column:   # column_ele is the position to be deleted
            del X_data[i][column_ele]

    return X_data

X_data = list(map(get_index, wordVectorSeqList))    # map: gain a element once
 
# Feature selection
X_data = feature_selection(X_data, useless_filename, xSquare_selection)

# padding
maxlen_tmp = len(X_data[0]) 
if maxlen_tmp < maxlen:
    maxlen = maxlen_tmp  

# Partition data set
# build the training set and test set of the first layer
X_pad_first = pad_sequences(X_data, maxlen=maxlen)
Y_first = wordVectorTarList   

X_train_first, Y_train_first, = X_pad_first, Y_first


# build the training set and test set of the second layer
X_pad_second = pad_sequences(X_data, maxlen=maxlen)     
Y_second = wordVectorStrengthList


X_train_second, Y_train_second = X_pad_second, Y_second


# Further build the training set required for the second layer model
final_X_train_second = [] # Sequence of training set
final_Y_train_second = [] # Label of the training set

for Seqindex in range(len(X_train_second)):  # train data set of 2nd
  if int(Y_train_second[Seqindex])==1 or int(Y_train_second[Seqindex])==0:
    final_X_train_second.append(X_train_second[Seqindex])
    final_Y_train_second.append(Y_train_second[Seqindex])
    
# The processed second layer training set
final_X_train_second = pad_sequences(final_X_train_second, maxlen=maxlen)


# Model framework
def Model_TransBERT(input_length, SeqPoseDict_length):
    '''
    Define the classification model
    '''
    INPUT_DIM = SeqPoseDict_length + 1
    OUTPUT_DIM = 768
    
    def attention_3d_block(inputs, TIME_STEPS):     # attention
        input_dim = int(inputs.shape[2])
        a = Permute((2, 1))(inputs)
        a = Reshape((input_dim, TIME_STEPS))(a) # this line is not useful. It's just to know which dimension is what.
        a = Dense(TIME_STEPS, activation='softmax')(a)        
        a_probs = Permute((2, 1), name='attention_vec')(a)
        output_attention_mul = Multiply()([inputs, a_probs])
        return output_attention_mul
        
    def create_lstm(input_length, rnn_hidden_dim=RNN_HIDDEN_DIM, output_dim=OUTPUT_DIM, input_dim=INPUT_DIM,
                    dropout_1=DROPOUT_RATIO_1, dropout_2=DROPOUT_RATIO_2):
        
        inputs_1 = Input(shape=(input_length, ))      
        inputs_2 = Embedding(input_dim=INPUT_DIM, output_dim=output_dim, name='embedding_layer',
                  trainable=True)(inputs_1)
        bi_lstm_1 = Bidirectional(LSTM(rnn_hidden_dim, return_sequences=True))(inputs_2)
        bn = BatchNormalization()(bi_lstm_1)
        attention_mul_bilstm = attention_3d_block(bn, input_length)
        drop_1 = Dropout(DROPOUT_RATIO_1)(attention_mul_bilstm)
        bi_lstm_1 = Bidirectional(LSTM(rnn_hidden_dim))(drop_1)
        drop_2 = Dropout(DROPOUT_RATIO_2)(bi_lstm_1)
        dense = Dense(1, activation='sigmoid')(drop_2)
        
        model = Model(inputs_1, dense)
        
        model.compile('adam', 'binary_crossentropy', metrics=['accuracy'])
                
        return model
    
    return create_lstm(input_length)


# First layer model
if __name__ == '__main__':
    model = Model_TransBERT(len(X_train_first[0]), SeqPoseDict_length)  # Get the first layer model

    saveBestModel = ModelCheckpoint(filepath, monitor='val_acc', verbose=1, save_best_only=True, mode='max')

    callbacks_list = [saveBestModel]

    print('Fitting first model...')
    history = model.fit(X_train_first, Y_train_first, batch_size=BATCH_SIZE, epochs=EPCOHS,
                        callbacks=callbacks_list, validation_split=0.1, verbose=1)  # Start training
    # serialize model to JSON
    model_json = model.to_json()  
    with open("model.json", "w") as json_file:
        json_file.write(model_json)  
    # serialize weights to HDF5
    model.save_weights("model.h5")  
    print("Saved model to disk")


# Second layer model
if __name__ == '__main__':
    model = Model_TransBERT(len(final_X_train_second[0]), SeqPoseDict_length)  # Get the second layer model

    saveBestModel = ModelCheckpoint(filepath_second, monitor='val_acc', verbose=1, save_best_only=True, mode='max')

    callbacks_list = [saveBestModel]
    
    print('Fitting second model...')
    history = model.fit(final_X_train_second, final_Y_train_second, batch_size=BATCH_SIZE, epochs=EPCOHS,
                        callbacks=callbacks_list, validation_split=0.1, verbose=1)  # Start training

    # serialize model to JSON
    model_json = model.to_json()  
    with open("model.json", "w") as json_file:
        json_file.write(model_json)  

    # serialize weights to HDF5
    model.save_weights("model.h5")  
    print("Saved model to disk")