import copy
import pandas as pd
import random
from sklearn.feature_selection import SelectKBest
from sklearn.feature_selection import chi2
import os


k_merList = [2]   # k-mer
kkk = k_merList[0]
xSquare_selection=45
vocab_index_number = 0  

output_dir = './output'  # Store model files
if not os.path.exists(output_dir):  # if the path does not exist, generate it
    os.mkdir(output_dir)

def read_data():
    finalFasta_positive = {}  
    fileList = ['./data/weak_enhancer.txt', './data/strong_enhancer.txt'] 
    for m in fileList:
        fasta = {}  
        with open(m) as file:
            sequence = ""
            for line in file:
                if line.startswith(">"):
                    name = line[1:].rstrip()
                    fasta[name] = ''
                    continue

                fasta[name] += line.rstrip() 

            finalFasta_positive = dict(finalFasta_positive, **fasta)

    finalFasta_negative = {}  
    fasta = {}  
    with open('./data/non-enhancer.txt') as file:
        sequence = ""
        for line in file:
            if line.startswith(">"):
                name = line[1:].rstrip()
                fasta[name] = ''
                continue
            fasta[name] += line.rstrip()  

        finalFasta_negative = dict(finalFasta_negative, **fasta)  

    AllPositive_items = finalFasta_positive.items()  
    AllNegative_items = finalFasta_negative.items() 
    sequenceList = []  
    for key, value in AllPositive_items:
        tmpDict = {'sequence': value, 'target': 1}  # Construct positive samples
        sequenceList.append(tmpDict)
    for key, value in AllNegative_items:
        tmpDict = {'sequence': value, 'target': 0}  # Construct negative samples
        sequenceList.append(tmpDict)  

    return  sequenceList

# Use kmer to segment raw data
def k_mer_split(sequenceList, k_merList):

    def k_mer(sequence, k):
        # kmer algorithm
        splitSequence = []  
        for i in range(len(sequence) - k + 1):
            tmp = ''
            for j in range(i, i + k):
                tmp += sequence[j]

            splitSequence.append(tmp)
        return splitSequence

    finalSequenceList = []
    for k_merNum in k_merList:  
        sequenceTmpList = copy.deepcopy(sequenceList)  
        for i in range(len(sequenceList)):  
            sequenceTmpList[i]['sequence'] = k_mer(sequenceList[i]['sequence'],
                                                   k_merNum)  
            finalSequenceList.append(sequenceTmpList[i]) 

    # Take each word as a feature,
    avertFinal = finalSequenceList[:len(sequenceList)]
    for i in range(len(k_merList)):
        tmplist = finalSequenceList[len(sequenceList) * i:(i + 1) * len(sequenceList)]
        for dictindex in range(len(tmplist)):
            avertFinal[dictindex]['sequence'] = avertFinal[dictindex]['sequence'] + tmplist[dictindex]['sequence']

    finalSequenceList = avertFinal # finalSequenceList is a list, each element of which is a dictionary, and the sequence in finalSequenceList has been split by k-mer

    wordVectorSeqList = []  
    wordVectorTarList = []  
    for allDictSeq in finalSequenceList:
        wordVectorSeqList.append(allDictSeq['sequence'])  
        wordVectorTarList.append(allDictSeq['target'])  

    return wordVectorSeqList, wordVectorTarList, len(wordVectorSeqList[0])

def gain_vocab_index(sequence_list, vocab_index, k_list):

    def specific_train_test(train_test_sequence, train_test_label, vocab_index):
        _sample = []  
        for label_index in range(len(train_test_label)):
            _sample.append(train_test_sequence[label_index][vocab_index])
        return _sample

    wordVectorSeqList, wordVectorTarList, total_feature_number = k_mer_split(sequence_list, k_list)  

    sample_feature = specific_train_test(wordVectorSeqList, wordVectorTarList, vocab_index)
    return  sample_feature, wordVectorTarList

sequenceList = read_data()  # Get raw data
random.seed(75)
random.shuffle(sequenceList)  # shuffle


random.seed(42)
sequenceList = random.sample(sequenceList, int(len(sequenceList)*0.9))

chi_list = []   # Record the chi-square value of each position
p_value_list = []  # Record the pvalue of each position
while vocab_index_number != 398:
    print('Current '+str(vocab_index_number)+'-th '+ 'position')
    sample_feature, Y_label = gain_vocab_index(sequenceList, vocab_index_number, k_merList)  

    X_total = pd.get_dummies(sample_feature, prefix='kmer')

    # Choose the K best SeqPose features
    model1 = SelectKBest(chi2, k='all')  
    model1.fit_transform(X_total, Y_label) 
    chi_tmp = model1.scores_  
    chi_list.append((sum(chi_tmp), vocab_index_number))      # Calculate the chi-squared corresponding to the current position

    P_value_tmp = model1.pvalues_  
    p_value_list.append((sum(P_value_tmp), vocab_index_number))   # Calculate the p value corresponding to the current position

    vocab_index_number += 1

sorted_index_Pvalue = sorted(p_value_list, key=lambda x:x[0])
sorted_index_chi = sorted(chi_list, key=lambda x:x[0])

SeqPose_features = []
for i in range(-xSquare_selection,0):
    SeqPose_features.append(sorted_index_Pvalue[i][1])
SeqPose_features.sort(reverse=True)      

my_df_1 = pd.DataFrame(SeqPose_features)  
my_df_1.to_csv('./output/Useless_SeqPose_features_'+str(kkk)+'_mer_'+str(xSquare_selection)+'.csv', index=False, header=False)